import ast
import functools

import pyparsing as pp
from pyparsing import ParseException

from rust2rpm import log, TARGET_ARCHES

pp.ParserElement.enablePackrat()

# ConfigurationPredicate :
#      ConfigurationOption
#    | ConfigurationAll
#    | ConfigurationAny
#    | ConfigurationNot
# ConfigurationOption :
#    IDENTIFIER (= (STRING_LITERAL | RAW_STRING_LITERAL))?
# ConfigurationAll
#    all ( ConfigurationPredicateList? )
# ConfigurationAny
#    any ( ConfigurationPredicateList? )
# ConfigurationNot
#    not ( ConfigurationPredicate )
# ConfigurationPredicateList
#    ConfigurationPredicate (, ConfigurationPredicate)* ,?

# cfg(target_os = "macos")
# cfg(any(foo, bar))
# cfg(all(unix, target_pointer_width = "32"))
# cfg(not(foo))

IDENT_CHARS = pp.alphas + "_", pp.alphanums + "_"


def _call(word, arg):
    return pp.Group(pp.Literal(word) + pp.Suppress("(") + arg + pp.Suppress(")"))


@functools.cache
def cfg_grammar():
    pred = pp.Forward()

    ident = pp.Word(IDENT_CHARS[0], IDENT_CHARS[1])
    option = pp.Group(ident + pp.Optional(pp.Suppress("=") + pp.quotedString))

    not_ = _call("not", pred)

    # pp.pyparsing_common.comma_separated_list?
    # any_ = _call('any', pp.pyparsing_common.comma_separated_list(pred))
    # all_ = _call('all', pp.pyparsing_common.comma_separated_list(pred))
    # all_ = _call('all', pp.delimited_list(pred))

    any_ = _call("any", pred + pp.ZeroOrMore(pp.Suppress(",") + pred))
    all_ = _call("all", pred + pp.ZeroOrMore(pp.Suppress(",") + pred))

    pred <<= not_ | any_ | all_ | option

    grammar = _call("cfg", pred)
    return grammar


@functools.cache
def evaluate_predicate(name: str, value: str, target_arch: str) -> bool:
    # based on: https://doc.rust-lang.org/reference/conditional-compilation.html

    if name == "target_arch":
            # Needs to be ignored, as we cannot generate patches that are
            # different depending on the host architecture - except if the
            # target architecture is "wasm32", which we don't support.
            return value == target_arch

    elif name == "target_feature":
            # The "target_feature" predicate can be ignored as well, since the
            # valid values for this predicate are architecture-dependent.
            return True

    elif name == "target_os":
            return value == "linux"

    elif name == "target_family":
            return value == "unix"

    elif name == "target_env":
            # The "target_env" predicate is used to disambiguate target
            # platforms based on its C library / C ABI (i.e. we can ignore
            # "msvc" and "musl"), and if there's no need to disambiguate, the
            # value can be the empty string.
            return value in ["", "gnu"]

    elif name == "target_endian":
            # Needs to be ignored, as we cannot generate patches that are
            # different depending on the host architecture.
            return True

    elif name == "target_pointer_width":
            # Needs to be ignored, as we cannot generate patches that are
            # different depending on the host architecture.
            return True

    elif name == "target_vendor":
            # On linux systems, "target_vendor" is always "unknown".
            return value == "unknown"

    else:
            log.warn(f'Ignoring invalid predicate \'"{name}" = "{value}"\' in cfg-expression.')
            return False


@functools.cache
def evaluate_atom(name: str) -> bool:
    if name == "unix":
            return True
    elif name == "windows":
            return False
    elif name == "miri":
            return False
    else:
            log.warn(
                f"Ignoring unknown identifier {name!r} in cfg-expression. "
                + 'only "unix", "windows", and "miri" are standard identifiers; '
                + 'any non-standard "--cfg" flags are ignored.'
            )
            return False


def evaluate(expr, target_arch: str, nested: bool = False) -> bool:
    if hasattr(expr, "asList"):
        expr = expr.asList()  # compat with pyparsing 2.7.x
    if not nested and len(expr) == 2 and expr[0] == "cfg":
            return evaluate(expr[1], target_arch, True)
    elif nested and len(expr) == 2 and expr[0] == "not":
            return not evaluate(expr[1], target_arch, True)
    elif nested and len(expr) >= 1 and expr[0] == "all":
            return all(evaluate(arg, target_arch, True) for arg in expr[1:])
    elif nested and len(expr) >= 1 and expr[0] == "any":
            return any(evaluate(arg, target_arch, True) for arg in expr[1:])
    elif nested and len(expr) == 2:
            v = ast.literal_eval(expr[1])
            return evaluate_predicate(expr[0], v, target_arch)
    elif nested and len(expr) == 1:
            return evaluate_atom(expr[0])
    else:
            raise ValueError


def parse_and_evaluate(expr):
    parsed = cfg_grammar().parseString(expr)[0]

    # evaluate cfg-expression for all supported target_arch values
    # returns True if it evaluates to True for any supported architecture
    return any(evaluate(parsed, target_arch) for target_arch in TARGET_ARCHES)
